from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.test import TestCase
from django.urls import reverse
from django.utils.translation import gettext as _

from mozilla_django_oidc_db.models import OpenIDConnectConfig

from .factories import StaffUserFactory

User = get_user_model()


class OIDCFlowTests(TestCase):
    @patch(
        "mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_or_create_user"
    )
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_userinfo")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.store_tokens")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_token")
    @patch(
        "mozilla_django_oidc_db.mixins.OpenIDConnectConfig.get_solo",
        return_value=OpenIDConnectConfig(enabled=True),
    )
    def test_duplicate_email_unique_constraint_violated(
        self,
        mock_get_solo,
        mock_get_token,
        mock_verify_token,
        mock_store_tokens,
        mock_get_userinfo,
        mock_get_or_create_user,
    ):
        """
        Assert that duplicate email addresses result in usable user feedback.
        """
        mock_get_or_create_user.side_effect = IntegrityError(
            """duplicate key value violates unique constraint "filled_email_unique"""
            """"\nDETAIL:  Key (email)=(collision@example.com) already exists.\n"""
        )
        # set up a user with a colliding email address
        mock_get_userinfo.return_value = {
            "email": "collision@example.com",
            "sub": "some_username",
        }
        StaffUserFactory.create(
            username="nonmatchingusername", email="collision@example.com"
        )
        session = self.client.session
        session["oidc_states"] = {"mock": {"nonce": "nonce"}}
        session.save()
        callback_url = reverse("oidc_authentication_callback")

        # enter the login flow
        callback_response = self.client.get(
            callback_url, {"code": "mock", "state": "mock"}
        )

        error_url = reverse("admin-oidc-error")

        with self.subTest("error redirects"):
            self.assertRedirects(callback_response, error_url)

        with self.subTest("exception info on error page"):
            error_page = self.client.get(error_url)

            self.assertEqual(error_page.status_code, 200)
            self.assertEqual(
                error_page.context["oidc_error"],
                """duplicate key value violates unique constraint "filled_email_unique"""
                """"\nDETAIL:  Key (email)=(collision@example.com) already exists.\n""",
            )
            self.assertContains(
                error_page, "duplicate key value violates unique constraint"
            )

    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_userinfo")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.store_tokens")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_token")
    @patch(
        "mozilla_django_oidc_db.mixins.OpenIDConnectConfig.get_solo",
        return_value=OpenIDConnectConfig(enabled=True),
    )
    def test_validation_error_during_authentication(
        self,
        mock_get_solo,
        mock_get_token,
        mock_verify_token,
        mock_store_tokens,
        mock_get_userinfo,
    ):
        """
        Assert that ValidationErrors raised during the auth process
        result in usable user feedback.
        """
        mock_get_solo.return_value = OpenIDConnectConfig(
            enabled=True,
            claim_mapping={
                "is_superuser": "missing_is_superuser",
                "email": "email",
                "first_name": "given_name",
                "last_name": "family_name",
            },
        )
        mock_get_token.return_value = {
            "id_token": "mock-id-token",
            "access_token": "mock-access-token",
        }
        # set up a user with the missing ``missing_is_superuser`` claim
        mock_get_userinfo.return_value = {
            "sub": "some_username",
            "email": "admin@example.com",
            "given_name": "John",
            "family_name": "Doe",
        }

        StaffUserFactory.create(username="some_username", email="admin@example.com")
        session = self.client.session
        session["oidc_states"] = {"mock": {"nonce": "nonce"}}
        session.save()
        callback_url = reverse("oidc_authentication_callback")

        # enter the login flow
        callback_response = self.client.get(
            callback_url, {"code": "mock", "state": "mock"}
        )

        error_url = reverse("admin-oidc-error")

        with self.subTest("error redirects"):
            self.assertRedirects(callback_response, error_url)

        with self.subTest("exception info on error page"):
            error_page = self.client.get(error_url)

            self.assertEqual(error_page.status_code, 200)
            err_msg = _("“%(value)s” value must be either True or False.") % {
                "value": ""
            }
            self.assertEqual(error_page.context["oidc_error"], err_msg)
            self.assertContains(error_page, err_msg)

    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_userinfo")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.store_tokens")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_token")
    @patch(
        "mozilla_django_oidc_db.mixins.OpenIDConnectConfig.get_solo",
        return_value=OpenIDConnectConfig(id=1, enabled=True),
    )
    def test_happy_flow(
        self,
        mock_get_solo,
        mock_get_token,
        mock_verify_token,
        mock_store_tokens,
        mock_get_userinfo,
    ):
        """
        Assert that duplicate email addresses result in usable user feedback.
        """
        # set up a user with a colliding email address
        mock_get_userinfo.return_value = {
            "email": "nocollision@example.com",
            "sub": "some_username",
        }
        StaffUserFactory.create(
            username="nonmatchingusername", email="collision@example.com"
        )
        session = self.client.session
        session["oidc_states"] = {"mock": {"nonce": "nonce"}}
        session.save()
        callback_url = reverse("oidc_authentication_callback")

        # enter the login flow
        callback_response = self.client.get(
            callback_url, {"code": "mock", "state": "mock"}
        )

        self.assertRedirects(
            callback_response, reverse("admin:index"), fetch_redirect_response=False
        )
        self.assertTrue(User.objects.filter(email="nocollision@example.com").exists())

    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_userinfo")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.store_tokens")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_token")
    @patch(
        "mozilla_django_oidc_db.mixins.OpenIDConnectConfig.get_solo",
        return_value=OpenIDConnectConfig(id=1, enabled=True),
    )
    def test_no_userinfo_defined(
        self,
        mock_get_solo,
        mock_get_token,
        mock_verify_token,
        mock_store_tokens,
        mock_get_userinfo,
    ):
        """
        Assert that the login procedure fails gracefully when no user claims are returned
        """
        # Testing with empty dict, because mozilla-django-oidc==1.2.x cannot deal with
        # `None` here: https://github.com/mozilla/mozilla-django-oidc/blob/1.2.4/mozilla_django_oidc/auth.py#L309
        mock_get_userinfo.return_value = {}
        StaffUserFactory.create(
            username="nonmatchingusername", email="collision@example.com"
        )
        session = self.client.session
        session["oidc_states"] = {"mock": {"nonce": "nonce"}}
        session.save()
        callback_url = reverse("oidc_authentication_callback")

        # enter the login flow
        callback_response = self.client.get(
            callback_url, {"code": "mock", "state": "mock"}
        )
        self.assertEqual(callback_response.status_code, 302)

        self.assertRedirects(
            callback_response,
            reverse("admin-oidc-error"),
            fetch_redirect_response=False,
        )
        self.assertFalse(User.objects.filter(email="nocollision@example.com").exists())

    def test_error_page_direct_access_forbidden(self):
        error_url = reverse("admin-oidc-error")

        response = self.client.get(error_url)

        self.assertEqual(response.status_code, 403)

    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_userinfo")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.store_tokens")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc_db.backends.OIDCAuthenticationBackend.get_token")
    @patch(
        "mozilla_django_oidc_db.mixins.OpenIDConnectConfig.get_solo",
        return_value=OpenIDConnectConfig(id=1, enabled=True),
    )
    def test_error_first_cleared_after_succesful_login(
        self,
        mock_get_solo,
        mock_get_token,
        mock_verify_token,
        mock_store_tokens,
        mock_get_userinfo,
    ):
        mock_get_userinfo.return_value = {
            "email": "nocollision@example.com",
            "sub": "some_username",
        }
        session = self.client.session
        session["oidc-error"] = "some error"
        session.save()
        error_url = reverse("admin-oidc-error")

        with self.subTest("with error"):
            response = self.client.get(error_url)

            self.assertEqual(response.status_code, 200)

        with self.subTest("after succesful login"):
            session["oidc_states"] = {"mock": {"nonce": "nonce"}}
            session.save()
            callback_url = reverse("oidc_authentication_callback")

            # enter the login flow
            callback_response = self.client.get(
                callback_url, {"code": "mock", "state": "mock"}
            )

            self.assertRedirects(
                callback_response, reverse("admin:index"), fetch_redirect_response=False
            )

            with self.subTest("check error page again"):
                response = self.client.get(error_url)

                self.assertEqual(response.status_code, 403)



from pyquery import PyQuery as pq
from requests import Session


def keycloak_login(
    login_url: str,
    username: str = "testuser",
    password: str = "testuser",
) -> str:
    """
    Test helper to perform a keycloak login.

    :param login_url: A login URL for keycloak with all query string parameters. E.g.
        `client.get(reverse("login"))["Location"]`.
    :returns: The redirect URI to consume in the django application, with the ``code``
        ``state`` query parameters. Consume this with ``response = client.get(url)``.
    """

    with Session() as session:
        login_page = session.get(login_url)
        assert login_page.status_code == 200

        # process keycloak's login form and submit the username + password to
        # authenticate
        document = pq(login_page.text)
        login_form = document("form#kc-form-login")
        submit_url = login_form.attr("action")
        assert isinstance(submit_url, str)
        login_response = session.post(
            submit_url,
            data={
                "username": username,
                "password": password,
                "credentialId": "",
                "login": "Sign In",
            },
            allow_redirects=False,
        )

        assert login_response.status_code == 302
        assert (redirect_uri := login_response.headers["Location"]).startswith(
            "http://testserver/"
        )

        return redirect_uri



# `classproperty` was moved to another module in Django 3.1
# See https://github.com/django/django/blob/ca9872905559026af82000e46cde6f7dedc897b6/docs/releases/3.1.txt#L649
try:
    from django.utils.functional import classproperty
except ImportError:
    from django.utils.decorators import classproperty



from typing import Iterator

import pytest

from mozilla_django_oidc_db.forms import OpenIDConnectConfigForm
from mozilla_django_oidc_db.models import OpenIDConnectConfig, get_default_scopes

KEYCLOAK_BASE_URL = "http://localhost:8080/realms/test/"


@pytest.fixture
def mock_state_and_nonce(mocker):
    mocker.patch(
        "mozilla_django_oidc.views.get_random_string",
        return_value="not-a-random-string",
    )


@pytest.fixture
def keycloak_config(db) -> Iterator[OpenIDConnectConfig]:
    """
    Keycloak configuration for the provided docker-compose.yml setup.

    This install a configuration (solo model) configured with the appropriate
    credentials.

    When not using VCR cassettes, make sure the service is up and running:

    .. code-block:: console

        docker-compose up -d

    """
    endpoints = OpenIDConnectConfigForm.get_endpoints_from_discovery(KEYCLOAK_BASE_URL)

    config, _ = OpenIDConnectConfig.objects.update_or_create(
        pk=OpenIDConnectConfig.singleton_instance_id,
        defaults={
            "enabled": True,
            "oidc_rp_client_id": "testid",
            "oidc_rp_client_secret": "7DB3KUAAizYCcmZufpHRVOcD0TOkNO3I",
            "oidc_rp_sign_algo": "RS256",
            **endpoints,
            "oidc_rp_scopes_list": get_default_scopes() + ["bsn", "kvk"],
            "sync_groups": False,
        },
    )
    # in case caching is setup, ensure that it is invalidated
    config.save()

    yield config

    OpenIDConnectConfig.clear_cache()



from unittest.mock import patch
from urllib.parse import parse_qs, urlparse

from django.contrib.sessions.middleware import SessionMiddleware
from django.http.response import HttpResponseRedirect
from django.test import RequestFactory

import pytest

from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OpenIDConnectConfig


def get_response(*args, **kwargs):
    pass


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_sessionrefresh_oidc_not_enabled(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=False,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
    )

    request = RequestFactory().get("/")

    # Running the middleware should return None, since OIDC is disabled
    result = SessionRefresh(get_response).process_request(request)

    assert result is None


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_sessionrefresh_config_always_refreshed(mock_get_solo):
    """
    Middleware should refresh the config on every call of `process_request`
    """
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
    )

    middleware = SessionRefresh(get_response)
    request = RequestFactory().get("/")
    SessionMiddleware(get_response).process_request(request)

    with patch(
        "mozilla_django_oidc_db.middleware.SessionRefresh.is_refreshable_url",
        return_value=True,
    ):
        with patch("mozilla_django_oidc.middleware.reverse", return_value="/callback"):
            result1 = middleware.process_request(request)

            # Update the config and call the middleware again (without reinstantiating)
            mock_get_solo.return_value = OpenIDConnectConfig(
                enabled=True,
                oidc_rp_client_id="some-other-id",
                oidc_rp_client_secret="secret",
                oidc_rp_sign_algo="HS256",
                oidc_rp_scopes_list=["openid", "email", "other_scope"],
                oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
                oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
                oidc_op_token_endpoint="http://some.endpoint/v1/token",
                oidc_op_user_endpoint="http://some.endpoint/v1/user",
            )
            result2 = middleware.process_request(request)

    assert isinstance(result1, HttpResponseRedirect)
    assert isinstance(result2, HttpResponseRedirect)

    parsed1 = parse_qs(urlparse(result1.url).query)
    assert parsed1["client_id"] == ["testid"]
    assert parsed1["scope"] == ["openid email"]

    parsed2 = parse_qs(urlparse(result2.url).query)
    assert parsed2["client_id"] == ["some-other-id"]
    assert parsed2["scope"] == ["openid email other_scope"]


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_sessionrefresh_config_use_defaults(mock_get_solo):
    """
    Middleware should use defaults from `mozilla-django-oidc`, for instance if
    `OIDC_AUTHENTICATION_CALLBACK_URL` is not explicity provided
    """
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
    )

    middleware = SessionRefresh(get_response)
    request = RequestFactory().get("/")
    SessionMiddleware(get_response).process_request(request)

    with patch(
        "mozilla_django_oidc_db.middleware.SessionRefresh.is_refreshable_url",
        return_value=True,
    ):
        result1 = middleware.process_request(request)

        # Update the config and call the middleware again (without reinstantiating)
        mock_get_solo.return_value = OpenIDConnectConfig(
            enabled=True,
            oidc_rp_client_id="some-other-id",
            oidc_rp_client_secret="secret",
            oidc_rp_sign_algo="HS256",
            oidc_rp_scopes_list=["openid", "email", "other_scope"],
            oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
            oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
            oidc_op_token_endpoint="http://some.endpoint/v1/token",
            oidc_op_user_endpoint="http://some.endpoint/v1/user",
        )
        result2 = middleware.process_request(request)

    assert isinstance(result1, HttpResponseRedirect)
    assert isinstance(result2, HttpResponseRedirect)

    parsed1 = parse_qs(urlparse(result1.url).query)
    assert parsed1["client_id"] == ["testid"]
    assert parsed1["scope"] == ["openid email"]

    parsed2 = parse_qs(urlparse(result2.url).query)
    assert parsed2["client_id"] == ["some-other-id"]
    assert parsed2["scope"] == ["openid email other_scope"]


@pytest.mark.django_db
def test_attributeerror_for_non_oidc_attribute():
    middleware = SessionRefresh(get_response)

    with pytest.raises(AttributeError):
        middleware.__name__

    # OIDC attributes should never raise AttributeErrors
    middleware.OIDC_AUTHENTICATION_CALLBACK_URL



from django.db import models
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import ArrayField


class ClaimField(ArrayField):
    """
    A field to store a path to claims holding the desired value(s).

    Each item is a segment in the path from the root to leaf for nested claims.
    """

    def __init__(self, *args, **kwargs):
        kwargs["base_field"] = models.CharField(_("claim path segment"), max_length=50)
        super().__init__(*args, **kwargs)



# Generated by Django 3.2.14 on 2022-08-03 08:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "0001_initial"),
        ("mozilla_django_oidc_db", "0008_auto_20220422_0849"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="default_groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="The default groups to which every user logging in with OIDC will be assigned",
                to="auth.Group",
                verbose_name="default groups",
            ),
        ),
    ]



from mozilla_django_oidc_db.utils import obfuscate_claim_value, obfuscate_claims


def test_obfuscate_string():
    value = "123456782"
    result = obfuscate_claim_value(value)

    assert result == "*******82"


def test_obfuscate_non_string():
    value = 12345
    result = obfuscate_claim_value(value)

    assert result == "****5"


def test_obfuscate_nested():
    claims = {
        "foo": "not_obfuscated",
        "foo.bar": "obfuscated",
        "some": {
            "nested": {
                "claim": "obfuscated",
                "claim2": "not_obfuscated",
            }
        },
        "object": {
            "foo": "obfuscated",
            "bar": "obfuscated",
        },
    }
    claims_to_obfuscate = [
        ["foo.bar"],
        ["some", "nested", "claim"],
        ["object"],
    ]
    expected_result = {
        "foo": "not_obfuscated",
        "foo.bar": "********ed",
        "some": {"nested": {"claim": "********ed", "claim2": "not_obfuscated"}},
        "object": {"foo": "********ed", "bar": "********ed"},
    }

    result = obfuscate_claims(claims, claims_to_obfuscate)

    assert result == expected_result



from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, path

from mozilla_django_oidc_db.views import (
    AdminLoginFailure,
    OIDCAuthenticationRequestView,
)

urlpatterns = [
    path("admin/login/failure/", AdminLoginFailure.as_view(), name="admin-oidc-error"),
    path("admin/", admin.site.urls),
    path("login", OIDCAuthenticationRequestView.as_view(), name="login"),
    path("oidc/", include("mozilla_django_oidc.urls")),
] + staticfiles_urlpatterns()



from unittest.mock import patch

import pytest

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OpenIDConnectConfig
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestView


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_settings_derived_from_model_oidc_not_enabled(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(enabled=False)

    backend = OIDCAuthenticationBackend()

    # verify that the settings are not set in __init__
    assert backend.OIDC_RP_CLIENT_ID is None
    assert backend.OIDC_RP_CLIENT_SECRET is None
    assert backend.OIDC_RP_SIGN_ALGO == "HS256"  # default from mozilla-django-oidc
    assert backend.OIDC_OP_JWKS_ENDPOINT is None
    assert backend.OIDC_OP_TOKEN_ENDPOINT is None
    assert backend.OIDC_OP_USER_ENDPOINT is None
    assert backend.OIDC_RP_IDP_SIGN_KEY is None


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_settings_derived_from_model_oidc_enabled(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
    )

    backend = OIDCAuthenticationBackend()

    # verify that the settings are derived from OpenIDConnectConfig
    assert backend.OIDC_RP_CLIENT_ID == "testid"
    assert backend.OIDC_RP_CLIENT_SECRET == "secret"
    assert backend.OIDC_RP_SIGN_ALGO == "HS256"
    assert backend.OIDC_OP_JWKS_ENDPOINT == "http://some.endpoint/v1/jwks"
    assert backend.OIDC_OP_TOKEN_ENDPOINT == "http://some.endpoint/v1/token"
    assert backend.OIDC_OP_USER_ENDPOINT == "http://some.endpoint/v1/user"
    assert backend.OIDC_RP_IDP_SIGN_KEY is None


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_view_settings_derived_from_model_oidc_enabled(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
    )

    view = OIDCAuthenticationRequestView()

    # verify that the settings are derived from OpenIDConnectConfig
    assert view.OIDC_RP_CLIENT_ID == "testid"
    assert view.OIDC_OP_AUTH_ENDPOINT == "http://some.endpoint/v1/auth"



"""
Support for user info JWT verification and decoding.

The bulk of the implementation is taken from mozilla-django-oidc where the access token
is processed, but adapted for non-hardcoded/configured parameters.

In the case of Keycloak for example, the token signing algorithm is configured on the
server and can change on a whim.
"""

import json
from typing import Any

from django.core.exceptions import SuspiciousOperation
from django.utils.encoding import smart_bytes

from josepy.jwk import JWK
from josepy.jws import JWS


def verify_and_decode_token(token: bytes, key) -> dict[str, Any]:
    """
    Verify that the token was not tampered with and if okay, return the payload.

    This is mostly taken from
    :meth:`mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws`.
    """

    jws = JWS.from_compact(token)

    # validate the signing algorithm
    if (alg := jws.signature.combined.alg) is None:
        raise SuspiciousOperation("No alg value found in header")

    # one of the most common implementation weaknesses -> attacker can supply 'none'
    # algorithm
    if alg.name == "none":
        raise SuspiciousOperation("'none' for alg value is not allowed")

    # process key parameter which was/may have been loaded from keys endpoint. The
    # string variant is unknown - this code is replicated from upstream
    # mozilla-django-oidc key verification.
    match key:
        case str():
            jwk = JWK.load(smart_bytes(key))
        case _:
            jwk = JWK.from_json(key)
            # address some missing upstream Self type declarations
            assert isinstance(jwk, JWK)

    if not jws.verify(jwk):
        raise SuspiciousOperation("JWS token verification failed.")

    return json.loads(jws.payload.decode("utf-8"))



from mozilla_django_oidc.middleware import SessionRefresh as _SessionRefresh

from .mixins import GetAttributeMixin, SoloConfigMixin


class SessionRefresh(GetAttributeMixin, SoloConfigMixin, _SessionRefresh):
    def __init__(self, get_response):
        # `super().__init__` is not called here, because this attempts to initialize
        # the settings (which should be retrieved from `OpenIDConnectConfig`).

        # The retrieval of these settings has been moved to runtime (`__getattribute__` from the `GetAttributeMixin`)
        super(_SessionRefresh, self).__init__(get_response=get_response)

    def process_request(self, request):
        # Initialize to retrieve the settings from config model
        super().__init__(self.get_response)

        self.refresh_config()
        if not self.config.enabled:
            return

        return super().process_request(request)



# Generated by Django 2.2.24 on 2021-11-29 10:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mozilla_django_oidc_db", "0005_openidconnectconfig_sync_groups_glob_pattern"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="username_claim",
            field=models.CharField(
                default="sub",
                help_text="The name of the OIDC claim that is used as the username",
                max_length=50,
                verbose_name="username claim",
            ),
        ),
    ]



# Generated by Django 3.2.23 on 2024-02-05 16:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0014_alter_openidconnectconfig_groups_claim"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="oidc_token_use_basic_auth",
            field=models.BooleanField(
                default=False,
                help_text="If enabled, the client ID and secret are sent in the HTTP Basic auth header when obtaining the access token. Otherwise, they are sent in the request body.",
                verbose_name="Use Basic auth for token endpoint",
            ),
        ),
    ]



import fnmatch
import logging
from typing import Any, TypeVar, cast

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, Group
from django.core.exceptions import ObjectDoesNotExist

import requests
from glom import Path, glom
from mozilla_django_oidc.auth import (
    OIDCAuthenticationBackend as _OIDCAuthenticationBackend,
)

from .jwt import verify_and_decode_token
from .mixins import GetAttributeMixin, SoloConfigMixin
from .models import OpenIDConnectConfig, UserInformationClaimsSources
from .utils import extract_content_type, obfuscate_claims

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=OpenIDConnectConfig)


class MissingIdentifierClaim(Exception):
    def __init__(self, claim_bits: list[str], *args, **kwargs):
        self.claim_bits = claim_bits
        super().__init__(*args, **kwargs)


class OIDCAuthenticationBackend(
    GetAttributeMixin, SoloConfigMixin[T], _OIDCAuthenticationBackend
):
    """
    Modifies the default OIDCAuthenticationBackend to use a configurable claim
    as unique identifier (default `sub`).
    """

    config_identifier_field = "username_claim"
    sensitive_claim_names: list[list[str]] = []

    def __init__(self, *args, **kwargs):
        # django-stubs returns AbstractBaseUser, but we depend on properties of
        # AbstractUser.
        self.UserModel = cast(AbstractUser, get_user_model())

        # See: https://github.com/maykinmedia/mozilla-django-oidc-db/issues/30
        # `super().__init__` is not called here, because this attempts to initialize
        # the settings (which should be retrieved from `OpenIDConnectConfig`).

        # The retrieval of these settings has been moved to runtime (`__getattribute__`)
        # to avoid a large number of `OpenIDConnectConfig.get_solo` calls when
        # `OIDCAuthenticationBackend.__init__` is called for permission checks

    def retrieve_identifier_claim(
        self, claims: dict, raise_on_empty: bool = False
    ) -> str:
        claim_bits = getattr(self.config, self.config_identifier_field)
        unique_id = glom(claims, Path(*claim_bits), default="")
        if raise_on_empty and not unique_id:
            raise MissingIdentifierClaim(claim_bits=claim_bits)
        return unique_id

    def get_sensitive_claims_names(self) -> list[list[str]]:
        """
        Defines the claims that should be obfuscated before logging claims.

        Nested claims are represented with a path of bits (e.g. ["foo", "bar", "baz"]).
        Claims with dots in them are supported, e.g. ["foo.bar"].
        """
        identifier_claim_bits: list[str] = getattr(
            self.config, self.config_identifier_field
        )
        return [identifier_claim_bits] + self.sensitive_claim_names

    def get_userinfo(self, access_token, id_token, payload):
        """
        Extract the user information, configurable whether to use the ID token or
        the userinfo endpoint for this
        """
        if self.config.userinfo_claims_source == UserInformationClaimsSources.id_token:
            logger.debug("Extracting user information from ID token")
            return payload

        logger.debug("Retrieving user information from userinfo endpoint")

        # copy of upstream get_userinfo which doesn't support application/jwt yet.
        # Overridden to handle application/jwt responses.
        # See https://github.com/mozilla/mozilla-django-oidc/issues/517
        #
        # Specifying the preferred format in the ``Accept`` header does not work with
        # Keycloak, as it depends on the client settings.
        user_response = requests.get(
            self.OIDC_OP_USER_ENDPOINT,
            headers={
                "Authorization": "Bearer {0}".format(access_token),
            },
            verify=self.get_settings("OIDC_VERIFY_SSL", True),
            timeout=self.get_settings("OIDC_TIMEOUT", None),
            proxies=self.get_settings("OIDC_PROXY", None),
        )
        user_response.raise_for_status()

        # From https://openid.net/specs/openid-connect-core-1_0.html#UserInfoResponse
        #
        # > The UserInfo Endpoint MUST return a content-type header to indicate which
        # > format is being returned.
        content_type = extract_content_type(user_response.headers["Content-Type"])
        match content_type:
            case "application/json":
                # the default case of upstream library
                return user_response.json()
            case "application/jwt":
                token = user_response.content
                # get the key from the configured keys endpoint
                # XXX: tested with asymmetric encryption. algorithms like HS256 rely on
                # out-of-band key exchange and are currently not supported until such a
                # case arrives.
                key = self.retrieve_matching_jwk(token)
                payload = verify_and_decode_token(token, key)
                return payload
            case _:
                raise ValueError(
                    f"Got an invalid Content-Type header value ({content_type}) "
                    "according to OpenID Connect Core 1.0 standard. Contact your "
                    "vendor."
                )

    def authenticate(self, *args, **kwargs):
        if not self.config.enabled:
            return None

        return super().authenticate(*args, **kwargs)

    def get_user_instance_values(self, claims) -> dict[str, Any]:
        """
        Map the names and values of the claims to the fields of the User model
        """
        return {
            model_field: glom(claims, Path(*claim_bits), default="")
            for model_field, claim_bits in self.config.claim_mapping.items()
        }

    def create_user(self, claims):
        """Return object for a newly created user account."""
        unique_id = self.retrieve_identifier_claim(claims)

        logger.debug("Creating OIDC user: %s", unique_id)

        user = self.UserModel.objects.create_user(
            **{self.UserModel.USERNAME_FIELD: unique_id}
        )
        self.update_user(user, claims)

        return user

    def filter_users_by_claims(self, claims):
        """Return all users matching the specified subject."""
        unique_id = self.retrieve_identifier_claim(claims)

        if not unique_id:
            return self.UserModel.objects.none()
        return self.UserModel.objects.filter(
            **{f"{self.UserModel.USERNAME_FIELD}__iexact": unique_id}
        )

    def verify_claims(self, claims) -> bool:
        """Verify the provided claims to decide if authentication should be allowed."""
        if not claims:
            return False

        claims_to_obfuscate = self.get_sensitive_claims_names()
        obfuscated_claims = obfuscate_claims(claims, claims_to_obfuscate)

        logger.debug("OIDC claims received: %s", obfuscated_claims)

        # check if we have an identifier
        try:
            self.retrieve_identifier_claim(claims, raise_on_empty=True)
        except MissingIdentifierClaim as exc:
            logger.error(
                "'%s' not in OIDC claims, cannot proceed with authentication",
                " > ".join(exc.claim_bits),
                exc_info=exc,
            )
            return False
        return True

    def update_user(self, user, claims):
        """Update existing user with new claims, if necessary save, and return user"""
        values = self.get_user_instance_values(claims)
        for field, value in values.items():
            setattr(user, field, value)
        logger.debug("Updating OIDC user %s with: %s", user, values)

        # Users can only be promoted to staff. Staff rights are never taken by OIDC.
        if self.config.make_users_staff and not user.is_staff:
            user.is_staff = True
            user.save(update_fields=["is_staff"])

        user.save(update_fields=values.keys())

        self.update_user_superuser_status(user, claims)

        self.update_user_groups(user, claims)
        self.update_user_default_groups(user)

        return user

    def _retrieve_groups_claim(self, claims: dict[str, Any]) -> list[str]:
        groups_claim_bits = self.config.groups_claim
        return glom(claims, Path(*groups_claim_bits), default=[])

    def update_user_superuser_status(self, user, claims) -> None:
        """
        Assigns superuser status to the user if the user is a member of at least one
        specific group. Superuser status is explicitly removed if the user is not or
        no longer member of at least one of these groups.
        """
        # can't do an isinstance check here
        superuser_group_names = cast(list[str], self.config.superuser_group_names)

        if not superuser_group_names:
            return

        claim_groups = self._retrieve_groups_claim(claims)
        if set(superuser_group_names) & set(claim_groups):
            user.is_superuser = True
        else:
            user.is_superuser = False
        user.save()

    def update_user_groups(self, user, claims) -> None:
        """
        Updates user group memberships based on the group_claim setting.

        Copied and modified from: https://github.com/snok/django-auth-adfs/blob/master/django_auth_adfs/backend.py
        """
        group_claim_bits: list[str] = self.config.groups_claim
        if not group_claim_bits:
            return

        claim_groups = self._retrieve_groups_claim(claims)

        # Update the user's group memberships
        django_groups = [group.name for group in user.groups.all()]
        if claim_groups:
            if not isinstance(claim_groups, list):
                claim_groups = [
                    claim_groups,
                ]
        else:
            logger.debug(
                "The configured groups claim '%s' was not found in the access token",
                " > ".join(group_claim_bits),
            )
            claim_groups = []
        if sorted(claim_groups) != sorted(django_groups):
            existing_groups = list(
                Group.objects.filter(name__in=claim_groups).iterator()
            )
            existing_group_names = frozenset(group.name for group in existing_groups)
            new_groups = []
            if self.config.sync_groups:
                # Only sync groups that match the supplied glob pattern
                new_groups = [
                    Group.objects.get_or_create(name=name)[0]
                    for name in fnmatch.filter(
                        claim_groups,
                        self.config.sync_groups_glob_pattern,
                    )
                    if name not in existing_group_names
                ]
            else:
                for name in claim_groups:
                    if name not in existing_group_names:
                        try:
                            group = Group.objects.get(name=name)
                            new_groups.append(group)
                        except ObjectDoesNotExist:
                            pass
            user.groups.set(existing_groups + new_groups)

    def update_user_default_groups(self, user):
        """
        Updates user group memberships based on the configured default groups.
        """
        default_groups = self.config.default_groups.all()
        user.groups.set(user.groups.union(default_groups))



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import ArrayField
from solo.models import SingletonModel, get_cache

import mozilla_django_oidc_db.settings as oidc_settings

from .compat import classproperty
from .fields import ClaimField


class UserInformationClaimsSources(models.TextChoices):
    userinfo_endpoint = "userinfo_endpoint", _("Userinfo endpoint")
    id_token = "id_token", _("ID token")


def get_default_scopes() -> list[str]:
    """
    Returns the default scopes to request for OpenID Connect logins
    """
    return ["openid", "email", "profile"]


def get_claim_mapping() -> dict[str, list[str]]:
    # Map (some) claim names from https://openid.net/specs/openid-connect-core-1_0.html#Claims
    # to corresponding field names on the User model
    return {
        "email": ["email"],
        "first_name": ["given_name"],
        "last_name": ["family_name"],
    }


def get_default_username_claim() -> list[str]:
    return ["sub"]


def get_default_groups_claim() -> list[str]:
    return ["roles"]


class CachingMixin:
    @classmethod
    def clear_cache(cls):
        cache_name = getattr(
            settings, "OIDC_CACHE", oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE
        )
        if cache_name:
            cache = get_cache(cache_name)
            cache_key = cls.get_cache_key()
            cache.delete(cache_key)

    def set_to_cache(self):
        cache_name = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_CACHE",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE,
        )
        if not cache_name:
            return None
        cache = get_cache(cache_name)
        cache_key = self.get_cache_key()
        timeout = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_CACHE_TIMEOUT",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE_TIMEOUT,
        )
        cache.set(cache_key, self, timeout)

    @classmethod
    def get_cache_key(cls) -> str:
        prefix = cls.custom_oidc_db_prefix or getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_PREFIX",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_PREFIX,
        )
        return "%s:%s" % (prefix, cls.__name__.lower())

    @classmethod
    def get_solo(cls) -> SingletonModel:
        cache_name = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_CACHE",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE,
        )
        if not cache_name:
            obj, created = cls.objects.get_or_create(pk=cls.singleton_instance_id)
            return obj
        cache = get_cache(cache_name)
        cache_key = cls.get_cache_key()
        obj = cache.get(cache_key)
        if not obj:
            obj, created = cls.objects.get_or_create(pk=cls.singleton_instance_id)
            obj.set_to_cache()
        return obj


class OpenIDConnectConfigBase(SingletonModel):
    """
    Defines the required fields for a config to establish an OIDC connection
    """

    enabled = models.BooleanField(
        _("enable"),
        default=False,
        help_text=_(
            "Indicates whether OpenID Connect for authentication/authorization is enabled"
        ),
    )

    oidc_rp_client_id = models.CharField(
        _("OpenID Connect client ID"),
        max_length=1000,
        help_text=_("OpenID Connect client ID provided by the OIDC Provider"),
    )
    oidc_rp_client_secret = models.CharField(
        _("OpenID Connect secret"),
        max_length=1000,
        help_text=_("OpenID Connect secret provided by the OIDC Provider"),
    )
    oidc_rp_sign_algo = models.CharField(
        _("OpenID sign algorithm"),
        max_length=50,
        help_text=_("Algorithm the Identity Provider uses to sign ID tokens"),
        default="HS256",
    )
    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("OpenID Connect scopes"),
        base_field=models.CharField(_("OpenID Connect scope"), max_length=50),
        default=get_default_scopes,
        blank=True,
        help_text=_("OpenID Connect scopes that are requested during login"),
    )

    oidc_op_discovery_endpoint = models.URLField(
        _("Discovery endpoint"),
        max_length=1000,
        help_text=_(
            "URL of your OpenID Connect provider discovery endpoint ending with a slash "
            "(`.well-known/...` will be added automatically). "
            "If this is provided, the remaining endpoints can be omitted, as "
            "they will be derived from this endpoint."
        ),
        blank=True,
    )
    oidc_op_jwks_endpoint = models.URLField(
        _("JSON Web Key Set endpoint"),
        max_length=1000,
        help_text=_(
            "URL of your OpenID Connect provider JSON Web Key Set endpoint. "
            "Required if `RS256` is used as signing algorithm."
        ),
        blank=True,
    )
    oidc_op_authorization_endpoint = models.URLField(
        _("Authorization endpoint"),
        max_length=1000,
        help_text=_("URL of your OpenID Connect provider authorization endpoint"),
    )
    oidc_op_token_endpoint = models.URLField(
        _("Token endpoint"),
        max_length=1000,
        help_text=_("URL of your OpenID Connect provider token endpoint"),
    )
    oidc_token_use_basic_auth = models.BooleanField(
        _("Use Basic auth for token endpoint"),
        default=False,
        help_text=_(
            "If enabled, the client ID and secret are sent in the HTTP Basic auth "
            "header when obtaining the access token. Otherwise, they are sent in the "
            "request body.",
        ),
    )
    oidc_op_user_endpoint = models.URLField(
        _("User endpoint"),
        max_length=1000,
        help_text=_("URL of your OpenID Connect provider userinfo endpoint"),
    )
    oidc_rp_idp_sign_key = models.CharField(
        _("Sign key"),
        max_length=1000,
        help_text=_(
            "Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. "
            "Should be the signing key in PEM or DER format."
        ),
        blank=True,
    )

    # Advanced settings
    oidc_use_nonce = models.BooleanField(
        _("Use nonce"),
        help_text=_(
            "Controls whether the OpenID Connect client uses nonce verification"
        ),
        default=True,
    )
    oidc_nonce_size = models.PositiveIntegerField(
        _("Nonce size"),
        help_text=_(
            "Sets the length of the random string used for OpenID Connect nonce verification"
        ),
        default=32,
    )
    oidc_state_size = models.PositiveIntegerField(
        _("State size"),
        help_text=_(
            "Sets the length of the random string used for OpenID Connect state verification"
        ),
        default=32,
    )
    oidc_exempt_urls = ArrayField(
        verbose_name=_("URLs exempt from session renewal"),
        base_field=models.CharField(_("Exempt URL"), max_length=1000),
        default=list,
        blank=True,
        help_text=_(
            "This is a list of absolute url paths, regular expressions for url paths, "
            "or Django view names. This plus the mozilla-django-oidc urls are exempted "
            "from the session renewal by the SessionRefresh middleware."
        ),
    )

    userinfo_claims_source = models.CharField(
        verbose_name=_("user information claims extracted from"),
        choices=UserInformationClaimsSources.choices,
        max_length=100,
        default=UserInformationClaimsSources.userinfo_endpoint,
        help_text=_(
            "Indicates the source from which the user information claims should be extracted."
        ),
    )

    @property
    def oidc_rp_scopes(self) -> str:
        """
        Scopes should be formatted as a string with spaces
        """
        return " ".join(self.oidc_rp_scopes_list)

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return force_str(self._meta.verbose_name)


class OpenIDConnectConfig(CachingMixin, OpenIDConnectConfigBase):
    """
    Configuration for authentication/authorization via OpenID connect
    """

    username_claim = ClaimField(
        verbose_name=_("username claim"),
        default=get_default_username_claim,
        help_text=_("The name of the OIDC claim that is used as the username"),
    )

    claim_mapping = models.JSONField(
        _("claim mapping"),
        default=get_claim_mapping,
        help_text=("Mapping from user-model fields to OIDC claims"),
    )
    groups_claim = ClaimField(
        verbose_name=_("groups claim"),
        default=get_default_groups_claim,
        help_text=_(
            "The name of the OIDC claim that holds the values to map to local user groups."
        ),
        blank=True,
    )

    sync_groups = models.BooleanField(
        _("Create local user groups if they do not exist yet"),
        default=True,
        help_text=_(
            "If checked, local user groups will be created for group names present in "
            "the groups claim, if they do not exist yet locally."
        ),
    )
    sync_groups_glob_pattern = models.CharField(
        _("groups glob pattern"),
        default="*",
        max_length=255,
        help_text=_(
            "The glob pattern that groups must match to be synchronized to "
            "the local database."
        ),
    )
    default_groups = models.ManyToManyField(
        Group,
        verbose_name=_("default groups"),
        blank=True,
        help_text=_(
            "The default groups to which every user logging in with OIDC will be assigned"
        ),
    )

    make_users_staff = models.BooleanField(
        _("make users staff"),
        default=False,
        help_text=_(
            "Users will be flagged as being a staff user automatically. This allows "
            "users to login to the admin interface. By default they have no permissions, even if they are staff."
        ),
    )
    superuser_group_names = ArrayField(
        verbose_name=_("Superuser group names"),
        base_field=models.CharField(_("Superuser group name"), max_length=50),
        default=list,
        blank=True,
        help_text=_(
            "If any of these group names are present in the claims upon login, "
            "the user will be marked as a superuser. If none of these groups are present "
            "the user will lose superuser permissions."
        ),
    )

    class Meta:
        verbose_name = _("OpenID Connect configuration")

    def clean(self):
        super().clean()

        # validate claim mapping
        User = get_user_model()
        for field in self.claim_mapping.keys():
            try:
                User._meta.get_field(field)
            except FieldDoesNotExist as exc:
                raise ValidationError(
                    {
                        "claim_mapping": _(
                            "Field '{field}' does not exist on the user model"
                        ).format(field=field)
                    }
                ) from exc

        if User.USERNAME_FIELD in self.claim_mapping:
            raise ValidationError(
                {
                    "claim_mapping": _(
                        "The username field may not be in the claim mapping"
                    ),
                }
            )

    @classproperty
    def custom_oidc_db_prefix(cls) -> str:
        """
        Cache prefix that can be overridden
        """
        return oidc_settings.MOZILLA_DJANGO_OIDC_DB_PREFIX



# Generated by Django 3.2.18 on 2023-12-21 14:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "mozilla_django_oidc_db",
            "0011_alter_openidconnectconfig_userinfo_claims_source",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="sync_groups",
            field=models.BooleanField(
                default=True,
                help_text="If checked, local user groups will be created for group names present in the groups claim, if they do not exist yet locally.",
                verbose_name="Create local user groups if they do not exist yet",
            ),
        ),
    ]



from copy import deepcopy
from typing import Any

from glom import Path, assign, glom
from requests.utils import _parse_content_type_header  # type: ignore


def obfuscate_claim_value(value: Any) -> str:
    """
    Obfuscates the value of a claim, so it can be logged safely

    If a dict of claims is supplied, all of the values will be obfuscated
    """
    if isinstance(value, dict):
        for k, v in value.items():
            value[k] = obfuscate_claim_value(v)
        return value
    else:
        value = str(value)
        threshold = int(len(value) * 0.75)
        return "".join([x if i > threshold else "*" for i, x in enumerate(value)])


def obfuscate_claims(claims: dict, claims_to_obfuscate: list[list[str]]) -> dict:
    """
    Obfuscates the specified claims in the specified claims dict
    """
    copied_claims = deepcopy(claims)
    for claim_bits in claims_to_obfuscate:
        claim_path = Path(*claim_bits)
        claim_value = glom(copied_claims, claim_path)
        assign(copied_claims, claim_path, obfuscate_claim_value(claim_value))
    return copied_claims


def extract_content_type(ct_header: str) -> str:
    """
    Get the content type + parameters from content type header.

    This is internal API since we use a requests internal utility, which may be
    removed/modified at any time. However, this is a deliberate choices since I trust
    requests to have a correct implementation more than coming up with one myself.
    """
    content_type, _ = _parse_content_type_header(ct_header)
    # discard the params, we only want the content type itself
    return content_type



import logging

from django.contrib import admin
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import IntegrityError, transaction
from django.urls import reverse_lazy
from django.views.generic import TemplateView

from mozilla_django_oidc.views import (
    OIDCAuthenticationCallbackView,
    OIDCAuthenticationRequestView as _OIDCAuthenticationRequestView,
)

from .mixins import SoloConfigMixin

logger = logging.getLogger(__name__)
OIDC_ERROR_SESSION_KEY = "oidc-error"


def get_exception_message(exc: Exception) -> str:
    if isinstance(exc, ValidationError):
        # ValidationError can be raised as part of django.db.models.fields.Field.to_python,
        # and unfortunately we don't have any context about the exact field that raised
        # the exception.
        return exc.messages[0]
    return exc.args[0]


class OIDCCallbackView(OIDCAuthenticationCallbackView):
    """
    Intercept errors raised by the authentication backend and display them.
    """

    failure_url = reverse_lazy("admin-oidc-error")

    def get(self, request):
        try:
            # ensure errors don't lead to half-created users
            with transaction.atomic():
                response = super().get(request)
        except (IntegrityError, ValidationError) as exc:
            logger.exception(
                "Something went wrong while attempting to authenticate via OIDC",
                exc_info=exc,
            )
            exc_message = get_exception_message(exc)
            request.session[OIDC_ERROR_SESSION_KEY] = exc_message
            return self.login_failure()
        else:
            if OIDC_ERROR_SESSION_KEY in request.session:
                del request.session[OIDC_ERROR_SESSION_KEY]
        return response


class AdminLoginFailure(TemplateView):
    """
    Template view in admin style to display OIDC login errors
    """

    template_name = "admin/oidc_failure.html"

    def dispatch(self, request, *args, **kwargs):
        if OIDC_ERROR_SESSION_KEY not in request.session:
            raise PermissionDenied()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(admin.site.each_context(self.request))
        context["oidc_error"] = self.request.session[OIDC_ERROR_SESSION_KEY]
        return context


class OIDCAuthenticationRequestView(SoloConfigMixin, _OIDCAuthenticationRequestView):
    pass



# Generated by Django 4.2.9 on 2024-05-01 16:10

from django.conf import settings
from django.core.cache import caches
from django.db import migrations, models, transaction

import mozilla_django_oidc_db.fields
import mozilla_django_oidc_db.models
import mozilla_django_oidc_db.settings as oidc_settings


def flush_cache():
    cache_name = getattr(
        settings,
        "MOZILLA_DJANGO_OIDC_DB_CACHE",
        oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE,
    )
    if not cache_name:
        return
    caches[cache_name].clear()


def forward(config) -> None:
    config.new_username_claim = config.username_claim.split(".")
    config.new_groups_claim = config.groups_claim.split(".")
    config.claim_mapping = {
        key: value.split(".") for key, value in config.claim_mapping.items()
    }


def reverse(config) -> None:
    config.username_claim = ".".join(config.new_username_claim)
    config.groups_claim = ".".join(config.new_groups_claim)
    config.claim_mapping = {
        key: ".".join(value) for key, value in config.claim_mapping.items()
    }


def action_factory(transformer):
    def _run_python_action(apps, _) -> None:
        OpenIDConnectConfig = apps.get_model(
            "mozilla_django_oidc_db", "OpenIDConnectConfig"
        )

        # Solo model, so there's only ever one instance
        config = OpenIDConnectConfig.objects.first()
        if config is None:
            return

        transformer(config)

        config.save()
        transaction.on_commit(flush_cache)

    return _run_python_action


copy_forward = action_factory(transformer=forward)
copy_reverse = action_factory(transformer=reverse)


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0001_initial_to_v015"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="new_groups_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                blank=True,
                default=mozilla_django_oidc_db.models.get_default_groups_claim,
                help_text="The name of the OIDC claim that holds the values to map to local user groups.",
                size=None,
                verbose_name="groups claim",
            ),
        ),
        migrations.AddField(
            model_name="openidconnectconfig",
            name="new_username_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.models.get_default_username_claim,
                help_text="The name of the OIDC claim that is used as the username",
                size=None,
                verbose_name="username claim",
            ),
        ),
        migrations.RunPython(copy_forward, copy_reverse),
        migrations.RemoveField(
            model_name="openidconnectconfig",
            name="groups_claim",
        ),
        migrations.RemoveField(
            model_name="openidconnectconfig",
            name="username_claim",
        ),
        migrations.RenameField(
            model_name="openidconnectconfig",
            old_name="new_groups_claim",
            new_name="groups_claim",
        ),
        migrations.RenameField(
            model_name="openidconnectconfig",
            old_name="new_username_claim",
            new_name="username_claim",
        ),
    ]



from typing import ClassVar, Generic, TypeVar, cast

from mozilla_django_oidc.utils import import_from_settings

from .models import OpenIDConnectConfig, OpenIDConnectConfigBase

T = TypeVar("T", bound=OpenIDConnectConfigBase)


class SoloConfigMixin(Generic[T]):
    config_class: ClassVar[type[OpenIDConnectConfigBase]] = OpenIDConnectConfig
    _solo_config: T

    @property
    def config(self) -> T:
        if not hasattr(self, "_solo_config"):
            # django-solo and type checking is challenging, but a new release is on the
            # way and should fix that :fingers_crossed:
            config = self.config_class.get_solo()
            self._solo_config = cast(T, config)
        return self._solo_config

    def refresh_config(self) -> None:
        """
        Refreshes the cached config on the instance, required for middleware
        since middleware is only instantiated once (during the Django startup phase)
        """
        if hasattr(self, "_solo_config"):
            del self._solo_config

    def get_settings(self, attr, *args):
        attr_lowercase = attr.lower()
        if hasattr(self.config, attr_lowercase):
            # Workaround for OIDC_RP_IDP_SIGN_KEY being an empty string by default.
            # mozilla-django-oidc explicitly checks if `OIDC_RP_IDP_SIGN_KEY` is not `None`
            # https://github.com/mozilla/mozilla-django-oidc/blob/master/mozilla_django_oidc/auth.py#L189
            value_from_config = getattr(self.config, attr_lowercase)
            if value_from_config == "":
                return None
            return value_from_config
        return import_from_settings(attr, *args)


class GetAttributeMixin:
    def __getattribute__(self, attr: str):
        """
        Mixin used to avoid calls to the config model on __init__ and instead
        do these calls runtime
        """
        if not attr.startswith("OIDC"):
            return super().__getattribute__(attr)

        try:
            default = super().__getattribute__(attr)
        except AttributeError:
            default = None
        return self.get_settings(attr, default)



from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.test import RequestFactory

import pytest

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.models import OpenIDConnectConfig


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_authenticate_oidc_not_enabled(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(enabled=False)

    backend = OIDCAuthenticationBackend()

    request = RequestFactory().get("/")

    # Authentication with the backend should not return a result,
    # because OIDC is not enabled
    assert backend.authenticate(request) is None


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_get_sensitive_claims(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True, username_claim=["sub"]
    )

    class CustomOIDCBackend(OIDCAuthenticationBackend):
        sensitive_claim_names = [["sensitive_claim1"], ["sensitive_claim2"]]

    backend = CustomOIDCBackend()

    # Only the sensitive claims + the identifier claim should be obfuscated
    assert backend.get_sensitive_claims_names() == [
        ["sub"],
        ["sensitive_claim1"],
        ["sensitive_claim2"],
    ]


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_get_user_instance_values(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        claim_mapping=OpenIDConnectConfig._meta.get_field("claim_mapping").get_default()
    )

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    user_values = backend.get_user_instance_values(claims)

    assert user_values == {
        "email": "admin@localhost",
        "first_name": "John",
        "last_name": "Doe",
    }


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_get_user_instance_values_nested_claims(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        claim_mapping={
            "email": ["user_info", "email"],
            "first_name": ["user_info", "given_name"],
            "last_name": ["user_info", "family_name"],
        }
    )

    claims = {
        "sub": "123456",
        "user_info": {
            "email": "admin@localhost",
            "given_name": "John",
            "family_name": "Doe",
        },
    }

    backend = OIDCAuthenticationBackend()

    user_values = backend.get_user_instance_values(claims)

    assert user_values == {
        "email": "admin@localhost",
        "first_name": "John",
        "last_name": "Doe",
    }


@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_supports_dots_in_claim_names(mock_get_solo, django_user_model):
    user = django_user_model.objects.create_user(username="dummy", password="dummy")
    mock_get_solo.return_value = OpenIDConnectConfig(
        username_claim=["ns1.sub"],
        groups_claim=["ns1.groups"],
        claim_mapping={
            "email": ["user_info.email"],
            "first_name": ["user_info.given_name"],
            "last_name": ["user_info.family_name"],
        },
    )

    claims = {
        "ns1.sub": "123456",
        "ns1.groups": ["aaaa"],
        "user_info.email": "admin@localhost",
        "user_info.given_name": "John",
        "user_info.family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    assert backend.retrieve_identifier_claim(claims) == "123456"
    assert backend.get_user_instance_values(claims) == {
        "email": "admin@localhost",
        "first_name": "John",
        "last_name": "Doe",
    }

    backend.update_user_groups(user, claims)
    group_names = user.groups.values_list("name", flat=True)
    assert list(group_names) == ["aaaa"]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
    )

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_different_username_claim(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        username_claim=["upn"],
    )

    claims = {
        "sub": "123456",
        "upn": "admin",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify `upn` is used as username
    assert user.username == "admin"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_filter_users(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        username_claim=["sub"],
    )

    User = get_user_model()

    # Create two users with the same email address, this shouldn't cause problems
    user1 = User.objects.create(
        username="123456", email="admin@localhost", first_name="John", last_name="Doe"
    )
    user2 = User.objects.create(
        username="654321", email="admin@localhost", first_name="Jane", last_name="Doe"
    )

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    users = backend.filter_users_by_claims(claims)

    assert users.count() == 1

    user = users.first()

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_filter_users_different_username_claim(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        username_claim=["upn"],
    )

    User = get_user_model()

    # Create two users with the same email address, this shouldn't cause problems
    user1 = User.objects.create(
        username="admin", email="admin@localhost", first_name="John", last_name="Doe"
    )
    user2 = User.objects.create(
        username="admin2", email="admin@localhost", first_name="Jane", last_name="Doe"
    )

    claims = {
        "upn": "admin2",
        "email": "admin@localhost",
        "given_name": "Jane",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    users = backend.filter_users_by_claims(claims)

    assert users.count() == 1

    user = users.first()

    # Verify that a user is created with the correct values
    assert user.username == "admin2"
    assert user.email == "admin@localhost"
    assert user.first_name == "Jane"
    assert user.last_name == "Doe"


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_update_user(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        username_claim=["sub"],
    )

    User = get_user_model()

    # Create two users with the same email address, this shouldn't cause problems
    # is_superuser should not be affected if `superuser_group_names` is not set
    user1 = User.objects.create(
        username="123456",
        email="admin@localhost",
        first_name="John",
        last_name="Doe",
        is_superuser=True,
    )
    user2 = User.objects.create(
        username="654321", email="admin@localhost", first_name="Jane", last_name="Doe"
    )

    claims = {
        "sub": "123456",
        "email": "modified@localhost",
        "given_name": "Name",
        "family_name": "Modified",
    }

    backend = OIDCAuthenticationBackend()

    user = backend.update_user(user1, claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "modified@localhost"
    assert user.first_name == "Name"
    assert user.last_name == "Modified"
    assert user.is_superuser


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_sync_all_groups(mock_get_solo):
    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=True,
        sync_groups_glob_pattern="*",
    )
    # Explicitly set to none
    oidc_config.default_groups.set(Group.objects.none())
    mock_get_solo.return_value = oidc_config

    claims = {
        "sub": "123456",
        "roles": ["useradmin", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that the groups were created
    assert Group.objects.count() == 2

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == [
        "useradmin",
        "groupadmin",
    ]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_no_groups_sync_without_groups_claim(mock_get_solo):
    Group.objects.create(name="group1")
    Group.objects.create(name="group2")

    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=[],
        sync_groups=True,
        sync_groups_glob_pattern="*",
    )
    # Explicitly set to none
    oidc_config.default_groups.set(Group.objects.none())
    mock_get_solo.return_value = oidc_config

    claims = {
        "sub": "123456",
        "roles": ["group1", "newgroup"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that no groups were created
    assert Group.objects.count() == 2

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == []


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_sync_groups_according_to_pattern(mock_get_solo):
    Group.objects.all().delete()

    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=True,
        sync_groups_glob_pattern="group*",
    )
    # Explicitly set to none
    oidc_config.default_groups.set(Group.objects.none())
    mock_get_solo.return_value = oidc_config

    claims = {
        "sub": "123456",
        "roles": ["useradmin", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == ["groupadmin"]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_sync_all_groups_nested_groups_claim(mock_get_solo):
    mock_get_solo.return_value = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["nested_object", "roles"],
        sync_groups=True,
        sync_groups_glob_pattern="*",
    )

    claims = {
        "sub": "123456",
        "nested_object": {"roles": ["useradmin", "groupadmin"]},
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that the groups were created
    assert Group.objects.count() == 2

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == [
        "useradmin",
        "groupadmin",
    ]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_sync_all_groups_and_default_groups(mock_get_solo):
    Group.objects.all().delete()

    default1 = Group.objects.create(name="default1")
    default2 = Group.objects.create(name="default2")
    Group.objects.create(name="other")

    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=True,
        sync_groups_glob_pattern="*",
    )
    oidc_config.save()
    oidc_config.default_groups.set([default1, default2])
    mock_get_solo.return_value = oidc_config

    claims = {
        "sub": "123456",
        "roles": ["useradmin", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == [
        "default1",
        "default2",
        "useradmin",
        "groupadmin",
    ]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_sync_groups_according_to_pattern_and_default_groups(
    mock_get_solo,
):
    Group.objects.all().delete()

    default1 = Group.objects.create(name="default1")
    default2 = Group.objects.create(name="default2")
    Group.objects.create(name="other")

    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=True,
        sync_groups_glob_pattern="group*",
    )
    oidc_config.save()
    oidc_config.default_groups.set([default1, default2])
    mock_get_solo.return_value = oidc_config

    claims = {
        "sub": "123456",
        "roles": ["useradmin", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == [
        "default1",
        "default2",
        "groupadmin",
    ]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_create_user_with_profile_settings(mock_get_solo):
    Group.objects.all().delete()

    mock_get_solo.return_value = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=True,
        claim_mapping={
            "first_name": ["given_name"],
            "last_name": ["family_name"],
            "email": ["email"],
            "is_superuser": ["is_god"],
        },
        sync_groups_glob_pattern="*",
        make_users_staff=True,
    )

    Group.objects.create(name="useradmin")
    Group.objects.create(name="groupadmin")

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
        "is_god": 1,
        "roles": [
            "useradmin",
        ],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"
    assert user.is_staff == True
    assert user.is_superuser == True
    assert list(user.groups.values_list("name", flat=True)) == ["useradmin"]


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_init_cache_not_called(mock_get_solo):
    """
    Regression test for https://github.com/maykinmedia/mozilla-django-oidc-db/issues/30
    """

    mock_get_solo.return_value = OpenIDConnectConfig(enabled=False)

    User = get_user_model()
    user = User.objects.create(
        username="123456", email="admin@localhost", first_name="John", last_name="Doe"
    )

    with patch(
        "mozilla_django_oidc_db.backends.OIDCAuthenticationBackend",
        side_effect=OIDCAuthenticationBackend,
    ) as mock_init:
        # `User.has_perm` should cause all backends to be instantiated
        user.has_perm("test")

        assert mock_init.call_count == 1

    # `OpenIDConnectConfig.get_solo` should not be called when initializing the backend
    assert mock_get_solo.call_count == 0


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_update_user_superuser(mock_get_solo):
    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=False,
        superuser_group_names=["superuser"],
    )
    mock_get_solo.return_value = oidc_config

    claims = {
        "sub": "123456",
        "roles": ["superuser", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that the groups were created
    assert Group.objects.count() == 0

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.is_superuser


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_update_user_remove_superuser(mock_get_solo):
    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=False,
        superuser_group_names=["superuser"],
    )
    mock_get_solo.return_value = oidc_config

    User = get_user_model()
    user = User.objects.create(
        username="123456",
        email="admin@localhost",
        first_name="John",
        last_name="Doe",
        is_superuser=True,
    )

    claims = {
        "sub": "123456",
        "roles": ["nosuperuser", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.update_user(user, claims)

    # Verify that the groups were created
    assert Group.objects.count() == 0

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert not user.is_superuser


@pytest.mark.django_db
@patch("mozilla_django_oidc_db.models.OpenIDConnectConfig.get_solo")
def test_backend_update_user_no_superuser_group_names(
    mock_get_solo,
):
    oidc_config = OpenIDConnectConfig(
        id=1,
        enabled=True,
        oidc_rp_client_id="testid",
        oidc_rp_client_secret="secret",
        oidc_rp_sign_algo="HS256",
        oidc_rp_scopes_list=["openid", "email"],
        oidc_op_jwks_endpoint="http://some.endpoint/v1/jwks",
        oidc_op_authorization_endpoint="http://some.endpoint/v1/auth",
        oidc_op_token_endpoint="http://some.endpoint/v1/token",
        oidc_op_user_endpoint="http://some.endpoint/v1/user",
        groups_claim=["roles"],
        sync_groups=False,
        superuser_group_names=[],
    )
    mock_get_solo.return_value = oidc_config

    User = get_user_model()
    user = User.objects.create(
        username="123456",
        email="admin@localhost",
        first_name="John",
        last_name="Doe",
        is_superuser=True,
    )

    claims = {
        "sub": "123456",
        "roles": ["nosuperuser", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.update_user(user, claims)

    # Verify that the groups were created
    assert Group.objects.count() == 0

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.is_superuser



from django.core.exceptions import ValidationError
from django.test.utils import isolate_apps
from django.utils.translation import gettext as _

import pytest

from mozilla_django_oidc_db.models import OpenIDConnectConfig, classproperty


def test_default_cache_key():
    assert OpenIDConnectConfig.get_cache_key() == "oidc:openidconnectconfig"


@isolate_apps("testapp")
def test_override_cache_key_with_class():
    class CustomConfig(OpenIDConnectConfig):
        class Meta:
            app_label = "custom"

        @classproperty
        def custom_oidc_db_prefix(cls):
            return "custom"

    assert CustomConfig.get_cache_key() == "custom:customconfig"


@isolate_apps("testapp")
def test_custom_config_override_cache_key_with_settings():
    class CustomConfig(OpenIDConnectConfig):
        class Meta:
            app_label = "custom"

        @classproperty
        def custom_oidc_db_prefix(cls):
            return ""

    # Prefix taken from `testapp/settings.py`
    assert CustomConfig.get_cache_key() == "default:customconfig"


@pytest.fixture()
def unset_oidc_cache_prefix(settings):
    del settings.MOZILLA_DJANGO_OIDC_DB_PREFIX


@isolate_apps("testapp")
def test_custom_config_cache_key_fallback(unset_oidc_cache_prefix):
    class CustomConfig(OpenIDConnectConfig):
        class Meta:
            app_label = "custom"

        @classproperty
        def custom_oidc_db_prefix(cls):
            return ""

    # Prefix taken from `mozilla_django_oidc_dv/settings.py`
    assert CustomConfig.get_cache_key() == "oidc:customconfig"


def test_validate_claim_mapping_fields():
    instance = OpenIDConnectConfig(
        claim_mapping={
            "bad_field_no_cookie": ["har"],
        }
    )

    with pytest.raises(ValidationError) as exc_context:
        instance.clean()

    err_dict = exc_context.value.message_dict
    assert "claim_mapping" in err_dict
    error = _("Field '{field}' does not exist on the user model").format(
        field="bad_field_no_cookie"
    )
    assert error in err_dict["claim_mapping"]


def test_validate_username_field_not_in_claim_mapping():
    instance = OpenIDConnectConfig(
        claim_mapping={
            "username": ["nope"],
        }
    )

    with pytest.raises(ValidationError) as exc_context:
        instance.clean()

    err_dict = exc_context.value.message_dict
    assert "claim_mapping" in err_dict
    error = _("The username field may not be in the claim mapping")
    assert error in err_dict["claim_mapping"]
